# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
from acrocord.misc import execution
import pytest


def test_deprecated():

    @execution.deprecated
    def foo():
        return "hello"

    with pytest.deprecated_call():
        foo()


def test_print():

    @execution.execution_time
    def foo():
        return "hello"

    assert foo() == "hello"
